# coding=utf8
"""Sopel GitHub

GitHub module for Sopel
"""
from __future__ import unicode_literals, absolute_import, division, print_function

from .github import *

__author__ = 'maxpowa'
__email__ = 'maxpowa@outlook.com'
__maintainer__ = 'dgw'
__version__ = '0.4.3'
__repo__ = 'https://github.com/sopel-irc/sopel-github'

